/*
 * Copyright (C) 2018-2023 Slava Monich <slava@monich.com>
 * Copyright (C) 2018-2019 Jolla Ltd.
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer
 *     in the documentation and/or other materials provided with the
 *     distribution.
 *
 *  3. Neither the names of the copyright holders nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * any official policies, either expressed or implied.
 */

#include "test_common.h"

#include "nfc_util.h"
#include "nfc_system.h"
#include "nfc_log.h"

static TestOpt test_opt;
static GString* test_log_buf;

static
void
test_log_proc(
    const char* name,
    int level,
    const char* format,
    va_list va)
{
    g_string_append_vprintf(test_log_buf, format, va);
    g_string_append(test_log_buf, "\n");
}

/*==========================================================================*
 * hexdump
 *==========================================================================*/

static
void
test_hexdump(
    void)
{
    static const guint8 data[] = {
        0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
        0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66
    };
    const char data_hexdump[] = "  0000: 30 31 32 33 34 35 36 37  "
        "38 39 61 62 63 64 65 66    01234567 89abcdef\n";
    const GLogProc fn = gutil_log_func;
    const int level = NFC_CORE_LOG_MODULE.level;
    GUtilData buf;

    test_log_buf = g_string_new(NULL);
    gutil_log_func = test_log_proc;

    /* No dump at this level */
    NFC_CORE_LOG_MODULE.level = GLOG_LEVEL_DEBUG;
    nfc_hexdump(data, sizeof(data));
    g_assert(!test_log_buf->len);

    /* Only at VERBOSE */
    NFC_CORE_LOG_MODULE.level = GLOG_LEVEL_VERBOSE;
    nfc_hexdump(data, sizeof(data));
    g_assert_cmpstr(test_log_buf->str, == ,data_hexdump);

    /* Same thing with GUtilData */
    g_string_set_size(test_log_buf, 0);
    TEST_BYTES_SET(buf, data);
    nfc_hexdump_data(NULL); /* This one does nothing */
    nfc_hexdump_data(&buf);
    g_assert_cmpstr(test_log_buf->str, == ,data_hexdump);

    NFC_CORE_LOG_MODULE.level = level;
    g_string_free(test_log_buf, TRUE);
    gutil_log_func = fn;
}

/*==========================================================================*
 * apdu/encode/fail
 *==========================================================================*/

static
void
test_apdu_encode_fail(
    void)
{
    static const NfcApdu apdu1 =
        { 0x01, 0x02, 0x03, 0x04, { NULL, 0x10000 }, 0 };
    static const NfcApdu apdu2 =
        { 0x01, 0x02, 0x03, 0x04, { NULL, 0 }, 0x10001 };
    GByteArray* buf = g_byte_array_new();

    g_assert(!nfc_apdu_encode(buf, &apdu1));
    g_assert_cmpuint(buf->len, == ,0);

    g_assert(!nfc_apdu_encode(buf, &apdu2));
    g_assert_cmpuint(buf->len, == ,0);

    g_byte_array_free(buf, TRUE);
}

/*==========================================================================*
 * apdu/decode/fail
 *==========================================================================*/

static
void
test_apdu_decode_fail(
    void)
{
    static const guint8 not_enough_data[] = {
        0x01, 0x02, 0x03
    };
    static const GUtilData not_enough = {
        TEST_ARRAY_AND_SIZE(not_enough_data)
    };
    static const guint8 too_much_1_data[] = {
        0x01, 0x02, 0x03, 0x04, 0x00
        /* nfc_apdu_decode doesn't read beyond that */
    };
    static const GUtilData too_much_1 = {
        too_much_1_data, 65545
    };
    static const guint8 too_much_2_data[] = {
        0x01, 0x02, 0x03, 0x04, 0xff
        /* nfc_apdu_decode doesn't read beyond that */
    };
    static const GUtilData too_much_2 = {
        too_much_2_data, 262
    };
    static const guint8 wrong_lc_1_data[] = {
        0x01, 0x02, 0x03, 0x04, 0x00, 0xcc, 0xcc, 0xdd
    };
    static const GUtilData wrong_lc_1 = {
        TEST_ARRAY_AND_SIZE(wrong_lc_1_data)
    };
    static const guint8 wrong_lc_2_data[] = {
        0x01, 0x02, 0x03, 0x04, 0xff, 0x01
    };
    static const GUtilData wrong_lc_2 = {
        TEST_ARRAY_AND_SIZE(wrong_lc_2_data)
    };
    NfcApdu apdu;

    g_assert(!nfc_apdu_decode(&apdu, &not_enough));
    g_assert(!nfc_apdu_decode(&apdu, &too_much_1));
    g_assert(!nfc_apdu_decode(&apdu, &too_much_2));
    g_assert(!nfc_apdu_decode(&apdu, &wrong_lc_1));
    g_assert(!nfc_apdu_decode(&apdu, &wrong_lc_2));
}

/*==========================================================================*
 * apdu/encode
 * apdu/decode
 *==========================================================================*/

typedef struct test_apdu_encode_data {
    const char* name;
    NfcApdu apdu;
    GUtilData encoded;
} TestApduEncodeData;

/* Case 1:  |CLA|INS|P1|P2|
 * Lc = 0, Le = 0, n = 4 */
static const guint8 test_apdu_encoded_case1[] = {
 /* CLA   INS   P1    P2 */
    0x01, 0x02, 0x03, 0x04
};

/* Case 2s: |CLA|INS|P1|P2|LE|
 * Lc = 0, Le = 1..256, n = 5) */
static const guint8 test_apdu_encoded_case2s_le_1[] = {
 /* CLA   INS   P1    P2    LE */
    0x01, 0x02, 0x03, 0x04, 0x01
};
static const guint8 test_apdu_encoded_case2s_le_255[] = {
    0x01, 0x02, 0x03, 0x04, 0xff
};
static const guint8 test_apdu_encoded_case2s_le_256[] = {
    0x01, 0x02, 0x03, 0x04, 0x00
};

/* Case 2e: |CLA|INS|P1|P2|00|LE1|LE2|
 * Lc = 0, Le = 257..65536, n = 7 */
static const guint8 test_apdu_encoded_case2e_le_257[] = {
 /* CLA   INS   P1    P2    00    LE1   LE2 */
    0x01, 0x02, 0x03, 0x04, 0x00, 0x01, 0x01
};
static const guint8 test_apdu_encoded_case2e_le_512[] = {
    0x01, 0x02, 0x03, 0x04, 0x00, 0x02, 0x00
};
static const guint8 test_apdu_encoded_case2e_le_65535[] = {
    0x01, 0x02, 0x03, 0x04, 0x00, 0xff, 0xff
};
static const guint8 test_apdu_encoded_case2e_le_65536[] = {
    0x01, 0x02, 0x03, 0x04, 0x00, 0x00, 0x00
};

/* Case 3s: |CLA|INS|P1|P2|LC|...BODY...|
 * Lc = 1..255, n = 6..260 */
static const guint8 test_apdu_encoded_case3s_lc_2[] = {
 /* CLA   INS   P1    P2    LC    BODY... */
    0x01, 0x02, 0x03, 0x04, 0x02, 0x05, 0x06
};
static const guint8 test_apdu_encoded_case3s_lc_255[] = {
    0x01, 0x02, 0x03, 0x04, 0xff, 0x00, 0x01, 0x02,
    0x03, 0x03, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a,
    0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12,
    0x13, 0x13, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a,
    0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20, 0x21, 0x22,
    0x23, 0x23, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a,
    0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32,
    0x33, 0x33, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a,
    0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42,
    0x43, 0x43, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a,
    0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50, 0x51, 0x52,
    0x53, 0x53, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a,
    0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60, 0x61, 0x62,
    0x63, 0x63, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a,
    0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72,
    0x73, 0x73, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a,
    0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80, 0x81, 0x82,
    0x83, 0x83, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a,
    0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90, 0x91, 0x92,
    0x93, 0x93, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a,
    0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 0xa0, 0xa1, 0xa2,
    0xa3, 0xa3, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa,
    0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2,
    0xb3, 0xb3, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba,
    0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2,
    0xc3, 0xc3, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca,
    0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd1, 0xd2,
    0xd3, 0xd3, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
    0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0, 0xe1, 0xe2,
    0xe3, 0xe3, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
    0xeb, 0xec, 0xed, 0xee, 0xef, 0xf0, 0xf1, 0xf2,
    0xf3, 0xf3, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa,
    0xfb, 0xfc, 0xfd, 0xfe
};

/* Case 3e: |CLA|INS|P1|P2|00|LC1|LC2|...BODY...|
 * Lc = 1..65535, n = 8..65542 */
static const guint8 test_apdu_encoded_case3e_lc_256[] = {
 /* CLA   INS   P1    P2    00    LC1   LC2  BODY... */
    0x01, 0x02, 0x03, 0x04, 0x00, 0x01, 0x00, 0x00,
    0x01, 0x02, 0x03, 0x03, 0x05, 0x06, 0x07, 0x08,
    0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10,
    0x11, 0x12, 0x13, 0x13, 0x15, 0x16, 0x17, 0x18,
    0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20,
    0x21, 0x22, 0x23, 0x23, 0x25, 0x26, 0x27, 0x28,
    0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30,
    0x31, 0x32, 0x33, 0x33, 0x35, 0x36, 0x37, 0x38,
    0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40,
    0x41, 0x42, 0x43, 0x43, 0x45, 0x46, 0x47, 0x48,
    0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50,
    0x51, 0x52, 0x53, 0x53, 0x55, 0x56, 0x57, 0x58,
    0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60,
    0x61, 0x62, 0x63, 0x63, 0x65, 0x66, 0x67, 0x68,
    0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70,
    0x71, 0x72, 0x73, 0x73, 0x75, 0x76, 0x77, 0x78,
    0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80,
    0x81, 0x82, 0x83, 0x83, 0x85, 0x86, 0x87, 0x88,
    0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90,
    0x91, 0x92, 0x93, 0x93, 0x95, 0x96, 0x97, 0x98,
    0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 0xa0,
    0xa1, 0xa2, 0xa3, 0xa3, 0xa5, 0xa6, 0xa7, 0xa8,
    0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0,
    0xb1, 0xb2, 0xb3, 0xb3, 0xb5, 0xb6, 0xb7, 0xb8,
    0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0,
    0xc1, 0xc2, 0xc3, 0xc3, 0xc5, 0xc6, 0xc7, 0xc8,
    0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0,
    0xd1, 0xd2, 0xd3, 0xd3, 0xd5, 0xd6, 0xd7, 0xd8,
    0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0,
    0xe1, 0xe2, 0xe3, 0xe3, 0xe5, 0xe6, 0xe7, 0xe8,
    0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 0xf0,
    0xf1, 0xf2, 0xf3, 0xf3, 0xf5, 0xf6, 0xf7, 0xf8,
    0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff
};

/* Case 4s: |CLA|INS|P1|P2|LC|...BODY...|LE |
 * Lc = 1..255, Le = 1..256, n = 7..261 */
static const guint8 test_apdu_encoded_case4s_lc_1_le_1[] = {
 /* CLA   INS   P1    P2    LC    BODY..LE */
    0x01, 0x02, 0x03, 0x04, 0x01, 0x05, 0x01
};
static const guint8 test_apdu_encoded_case4s_lc_2_le_256[] = {
    0x01, 0x02, 0x03, 0x04, 0x02, 0xdd, 0xdd, 0x00
};
static const guint8 test_apdu_encoded_case4s_lc_255_le_2[] = {
    0x01, 0x02, 0x03, 0x04, 0xff, 0x00, 0x01, 0x02,
    0x03, 0x03, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a,
    0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12,
    0x13, 0x13, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a,
    0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20, 0x21, 0x22,
    0x23, 0x23, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a,
    0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32,
    0x33, 0x33, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a,
    0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42,
    0x43, 0x43, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a,
    0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50, 0x51, 0x52,
    0x53, 0x53, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a,
    0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60, 0x61, 0x62,
    0x63, 0x63, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a,
    0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72,
    0x73, 0x73, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a,
    0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80, 0x81, 0x82,
    0x83, 0x83, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a,
    0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90, 0x91, 0x92,
    0x93, 0x93, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a,
    0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 0xa0, 0xa1, 0xa2,
    0xa3, 0xa3, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa,
    0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2,
    0xb3, 0xb3, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba,
    0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2,
    0xc3, 0xc3, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca,
    0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd1, 0xd2,
    0xd3, 0xd3, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
    0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0, 0xe1, 0xe2,
    0xe3, 0xe3, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
    0xeb, 0xec, 0xed, 0xee, 0xef, 0xf0, 0xf1, 0xf2,
    0xf3, 0xf3, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa,
    0xfb, 0xfc, 0xfd, 0xfe, 0x02 /* Le */
};

/* Case 4e: |CLA|INS|P1|P2|00|LC1|LC2|...BODY...|LE1|LE2|
 * Lc = 1..65535, Le = 1..65536, n = 10..65544 */
static const guint8 test_apdu_encoded_case4e_lc_256_le_1[] = {
 /* CLA   INS   P1    P2    00    LC1   LC2   BODY...  LE1 LE2 */
    0x01, 0x02, 0x03, 0x04, 0x00, 0x01, 0x00, 0x00,
    0x01, 0x02, 0x03, 0x03, 0x05, 0x06, 0x07, 0x08,
    0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10,
    0x11, 0x12, 0x13, 0x13, 0x15, 0x16, 0x17, 0x18,
    0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20,
    0x21, 0x22, 0x23, 0x23, 0x25, 0x26, 0x27, 0x28,
    0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30,
    0x31, 0x32, 0x33, 0x33, 0x35, 0x36, 0x37, 0x38,
    0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40,
    0x41, 0x42, 0x43, 0x43, 0x45, 0x46, 0x47, 0x48,
    0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50,
    0x51, 0x52, 0x53, 0x53, 0x55, 0x56, 0x57, 0x58,
    0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60,
    0x61, 0x62, 0x63, 0x63, 0x65, 0x66, 0x67, 0x68,
    0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70,
    0x71, 0x72, 0x73, 0x73, 0x75, 0x76, 0x77, 0x78,
    0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80,
    0x81, 0x82, 0x83, 0x83, 0x85, 0x86, 0x87, 0x88,
    0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90,
    0x91, 0x92, 0x93, 0x93, 0x95, 0x96, 0x97, 0x98,
    0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 0xa0,
    0xa1, 0xa2, 0xa3, 0xa3, 0xa5, 0xa6, 0xa7, 0xa8,
    0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0,
    0xb1, 0xb2, 0xb3, 0xb3, 0xb5, 0xb6, 0xb7, 0xb8,
    0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0,
    0xc1, 0xc2, 0xc3, 0xc3, 0xc5, 0xc6, 0xc7, 0xc8,
    0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0,
    0xd1, 0xd2, 0xd3, 0xd3, 0xd5, 0xd6, 0xd7, 0xd8,
    0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0,
    0xe1, 0xe2, 0xe3, 0xe3, 0xe5, 0xe6, 0xe7, 0xe8,
    0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 0xf0,
    0xf1, 0xf2, 0xf3, 0xf3, 0xf5, 0xf6, 0xf7, 0xf8,
    0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff, 0x00,
    0x01 /* Le */
};
static const guint8 test_apdu_encoded_case4e_lc_256_le_256[] = {
    0x01, 0x02, 0x03, 0x04, 0x00, 0x01, 0x00, 0x00,
    0x01, 0x02, 0x03, 0x03, 0x05, 0x06, 0x07, 0x08,
    0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10,
    0x11, 0x12, 0x13, 0x13, 0x15, 0x16, 0x17, 0x18,
    0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20,
    0x21, 0x22, 0x23, 0x23, 0x25, 0x26, 0x27, 0x28,
    0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30,
    0x31, 0x32, 0x33, 0x33, 0x35, 0x36, 0x37, 0x38,
    0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40,
    0x41, 0x42, 0x43, 0x43, 0x45, 0x46, 0x47, 0x48,
    0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50,
    0x51, 0x52, 0x53, 0x53, 0x55, 0x56, 0x57, 0x58,
    0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60,
    0x61, 0x62, 0x63, 0x63, 0x65, 0x66, 0x67, 0x68,
    0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70,
    0x71, 0x72, 0x73, 0x73, 0x75, 0x76, 0x77, 0x78,
    0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80,
    0x81, 0x82, 0x83, 0x83, 0x85, 0x86, 0x87, 0x88,
    0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90,
    0x91, 0x92, 0x93, 0x93, 0x95, 0x96, 0x97, 0x98,
    0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 0xa0,
    0xa1, 0xa2, 0xa3, 0xa3, 0xa5, 0xa6, 0xa7, 0xa8,
    0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0,
    0xb1, 0xb2, 0xb3, 0xb3, 0xb5, 0xb6, 0xb7, 0xb8,
    0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0,
    0xc1, 0xc2, 0xc3, 0xc3, 0xc5, 0xc6, 0xc7, 0xc8,
    0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0,
    0xd1, 0xd2, 0xd3, 0xd3, 0xd5, 0xd6, 0xd7, 0xd8,
    0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0,
    0xe1, 0xe2, 0xe3, 0xe3, 0xe5, 0xe6, 0xe7, 0xe8,
    0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 0xf0,
    0xf1, 0xf2, 0xf3, 0xf3, 0xf5, 0xf6, 0xf7, 0xf8,
    0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff, 0x01,
    0x00 /* Le */
};
static const guint8 test_apdu_encoded_case4e_lc_256_le_65536[] = {
 /* CLA   INS   P1    P2    00    LC1   LC2   BODY...  LE1 LE2 */
    0x01, 0x02, 0x03, 0x04, 0x00, 0x01, 0x00, 0x00,
    0x01, 0x02, 0x03, 0x03, 0x05, 0x06, 0x07, 0x08,
    0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10,
    0x11, 0x12, 0x13, 0x13, 0x15, 0x16, 0x17, 0x18,
    0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20,
    0x21, 0x22, 0x23, 0x23, 0x25, 0x26, 0x27, 0x28,
    0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30,
    0x31, 0x32, 0x33, 0x33, 0x35, 0x36, 0x37, 0x38,
    0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40,
    0x41, 0x42, 0x43, 0x43, 0x45, 0x46, 0x47, 0x48,
    0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50,
    0x51, 0x52, 0x53, 0x53, 0x55, 0x56, 0x57, 0x58,
    0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60,
    0x61, 0x62, 0x63, 0x63, 0x65, 0x66, 0x67, 0x68,
    0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70,
    0x71, 0x72, 0x73, 0x73, 0x75, 0x76, 0x77, 0x78,
    0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80,
    0x81, 0x82, 0x83, 0x83, 0x85, 0x86, 0x87, 0x88,
    0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90,
    0x91, 0x92, 0x93, 0x93, 0x95, 0x96, 0x97, 0x98,
    0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 0xa0,
    0xa1, 0xa2, 0xa3, 0xa3, 0xa5, 0xa6, 0xa7, 0xa8,
    0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0,
    0xb1, 0xb2, 0xb3, 0xb3, 0xb5, 0xb6, 0xb7, 0xb8,
    0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0,
    0xc1, 0xc2, 0xc3, 0xc3, 0xc5, 0xc6, 0xc7, 0xc8,
    0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0,
    0xd1, 0xd2, 0xd3, 0xd3, 0xd5, 0xd6, 0xd7, 0xd8,
    0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0,
    0xe1, 0xe2, 0xe3, 0xe3, 0xe5, 0xe6, 0xe7, 0xe8,
    0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 0xf0,
    0xf1, 0xf2, 0xf3, 0xf3, 0xf5, 0xf6, 0xf7, 0xf8,
    0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff, 0x00,
    0x00 /* Le */
};

static const TestApduEncodeData tests_apdu_encode[] = {
    /* Case 1:  |CLA|INS|P1|P2|
     * Lc = 0, Le = 0, n = 4 */
    {
        "case1",
        { 0x01, 0x02, 0x03, 0x04, { NULL, 0 }, 0 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case1) }
    },
    /* Case 2s: |CLA|INS|P1|P2|LE|
     * Lc = 0, Le = 1..256, n = 5) */
    {
        "case2s/le_1",
        { 0x01, 0x02, 0x03, 0x04, { NULL, 0 }, 1 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case2s_le_1) }
    },
    {
        "case2s/le_255",
        { 0x01, 0x02, 0x03, 0x04, { NULL, 0 }, 255 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case2s_le_255) }
    },
    {
        "case2s/le_256",
        { 0x01, 0x02, 0x03, 0x04, { NULL, 0 }, 256 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case2s_le_256) }
    },
    /* Case 2e: |CLA|INS|P1|P2|00|LE1|LE2|
     * Lc = 0, Le = 257..65536, n = 7 */
    {
        "case2e/le_257",
        { 0x01, 0x02, 0x03, 0x04, { NULL, 0 }, 257 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case2e_le_257) }
    },
    {
        "case2e/le_512",
        { 0x01, 0x02, 0x03, 0x04, { NULL, 0 }, 512 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case2e_le_512) }
    },
    {
        "case2e/le_65535",
        { 0x01, 0x02, 0x03, 0x04, { NULL, 0 }, 65535 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case2e_le_65535) }
    },
    {
        "case2e/le_65536",
        { 0x01, 0x02, 0x03, 0x04, { NULL, 0 }, 65536 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case2e_le_65536) }
    },
    /* Case 3s: |CLA|INS|P1|P2|LC|...BODY...|
     * Lc = 1..255, n = 6..260 */
    {
        "case3s/lc_2",
        { 0x01, 0x02, 0x03, 0x04,
        { test_apdu_encoded_case3s_lc_2 + 5, 2 }, 0 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case3s_lc_2) }
    },
    {
        "case3s/lc_255",
        { 0x01, 0x02, 0x03, 0x04,
        { test_apdu_encoded_case3s_lc_255 + 5, 255 }, 0 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case3s_lc_255) }
    },
    /* Case 3e: |CLA|INS|P1|P2|00|LC1|LC2|...BODY...|
     * Lc = 1..65535, n = 8..65542 */
    {
        "case3e/lc_256",
        { 0x01, 0x02, 0x03, 0x04,
        { test_apdu_encoded_case3e_lc_256 + 7, 256 }, 0 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case3e_lc_256) }
    },
    /* Case 4s: |CLA|INS|P1|P2|LC|...BODY...|LE |
     * Lc = 1..255, Le = 1..256, n = 7..261 */
    {
        "case4s/lc_1_le_1",
        { 0x01, 0x02, 0x03, 0x04,
        { test_apdu_encoded_case4s_lc_1_le_1 + 5, 1 }, 1 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case4s_lc_1_le_1) }
    },
    {
        "case4s/lc_2_le_256",
        { 0x01, 0x02, 0x03, 0x04,
        { test_apdu_encoded_case4s_lc_2_le_256 + 5, 2 }, 256 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case4s_lc_2_le_256) }
    },
    {
        "case4s/lc_255_le_2",
        { 0x01, 0x02, 0x03, 0x04,
        { test_apdu_encoded_case4s_lc_255_le_2 + 5, 255 }, 2 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case4s_lc_255_le_2) }
    },
    /* Case 4e: |CLA|INS|P1|P2|00|LC1|LC2|...BODY...|LE1|LE2|
     * Lc = 1..65535, Le = 1..65536, n = 10..65544 */
    {
        "case4e/lc_256_le_1",
        { 0x01, 0x02, 0x03, 0x04,
        { test_apdu_encoded_case4e_lc_256_le_1 + 7, 256 }, 1 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case4e_lc_256_le_1) }
    },
    {
        "case4e/lc_256_le_256",
        { 0x01, 0x02, 0x03, 0x04,
        { test_apdu_encoded_case4e_lc_256_le_256 + 7, 256 }, 256 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case4e_lc_256_le_256) }
    },
    {
        "case4e/lc_256_le_65356",
        { 0x01, 0x02, 0x03, 0x04,
        { test_apdu_encoded_case4e_lc_256_le_65536 + 7, 256 }, 65536 },
        { TEST_ARRAY_AND_SIZE(test_apdu_encoded_case4e_lc_256_le_65536) }
    }
};

static
void
test_apdu_encode(
    gconstpointer data)
{
    const TestApduEncodeData* test = data;
    GByteArray* buf = g_byte_array_new();

    g_assert(nfc_apdu_encode(buf, &test->apdu));
    g_assert_cmpuint(buf->len, == ,test->encoded.size);
    g_assert(!memcmp(buf->data, test->encoded.bytes, test->encoded.size));
    g_byte_array_free(buf, TRUE);
}

static
void
test_apdu_decode(
    gconstpointer data)
{
    const TestApduEncodeData* test = data;
    NfcApdu apdu;

    memset(&apdu, 0xaa, sizeof(apdu));
    g_assert(nfc_apdu_decode(&apdu, &test->encoded));
    g_assert_cmpuint(apdu.cla, == ,test->apdu.cla);
    g_assert_cmpuint(apdu.ins, == ,test->apdu.ins);
    g_assert_cmpuint(apdu.p1, == ,test->apdu.p1);
    g_assert_cmpuint(apdu.p2, == ,test->apdu.p2);
    g_assert_cmpuint(apdu.le, == ,test->apdu.le);
    g_assert_cmpuint(apdu.data.size, == ,test->apdu.data.size);
    if (test->apdu.data.size) {
        g_assert(!memcmp(apdu.data.bytes, test->apdu.data.bytes,
            apdu.data.size));
    } else {
        g_assert(!apdu.data.bytes);
    }
}

/*==========================================================================*
 * Common
 *==========================================================================*/

#define TEST_(name) "/core/util/" name

int main(int argc, char* argv[])
{
    guint i;

    g_test_init(&argc, &argv, NULL);
    g_test_add_func(TEST_("hexdump"), test_hexdump);
    g_test_add_func(TEST_("apdu/encode/fail"), test_apdu_encode_fail);
    g_test_add_func(TEST_("apdu/decode/fail"), test_apdu_decode_fail);
    for (i = 0; i < G_N_ELEMENTS(tests_apdu_encode); i++) {
        const TestApduEncodeData* test = tests_apdu_encode + i;
        char* path1 = g_strconcat(TEST_("apdu/encode/"), test->name, NULL);
        char* path2 = g_strconcat(TEST_("apdu/decode/"), test->name, NULL);

        g_test_add_data_func(path1, test, test_apdu_encode);
        g_test_add_data_func(path2, test, test_apdu_decode);
        g_free(path1);
        g_free(path2);
    }
    test_init(&test_opt, argc, argv);
    return g_test_run();
}

/*
 * Local Variables:
 * mode: C
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
